@@ -9,15 +9,14 @@
 -- Operations on concrete values
 -----------------------------------------------------------------------------
 
+{-# LANGUAGE DeriveDataTypeable  #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 {-# LANGUAGE TypeApplications    #-}
 {-# LANGUAGE Rank2Types          #-}
 
 {-# OPTIONS_GHC -Wall -Werror #-}
 
-module Data.SBV.Core.Concrete
-  ( module Data.SBV.Core.Concrete
-  ) where
+module Data.SBV.Core.Concrete where
 
 import Control.Monad (replicateM)
 
@@ -25,7 +24,7 @@ import Data.Bits
 import System.Random (randomIO, randomRIO)
 
 import Data.Char (chr, isSpace)
-import Data.List (isPrefixOf, intercalate)
+import Data.List (intercalate)
 
 import Data.SBV.Core.Kind
 import Data.SBV.Core.AlgReals
@@ -38,9 +37,12 @@ import Data.SBV.Utils.Numeric (fpIsEqualObjectH, fpCompareObjectH)
 import Data.Set (Set)
 import qualified Data.Set as Set
 
+import qualified Data.Generics as G
+
 -- | A 'RCSet' is either a regular set or a set given by its complement from the corresponding universal set.
 data RCSet a = RegularSet    (Set a)
              | ComplementSet (Set a)
+             deriving G.Data
 
 -- | Show instance. Regular sets are shown as usual.
 -- Complements are shown "U -" notation.
@@ -85,6 +87,7 @@ data CVal = CAlgReal  !AlgReal             -- ^ Algebraic real
           | CTuple    ![CVal]              -- ^ Tuple
           | CMaybe    !(Maybe CVal)        -- ^ Maybe
           | CEither   !(Either CVal CVal)  -- ^ Disjoint union
+          deriving G.Data
 
 -- | Assign a rank to constant values, this is structural and helps with ordering
 cvRank :: CVal -> Int
@@ -163,7 +166,7 @@ instance Ord CVal where
 data CV = CV { _cvKind  :: !Kind
              , cvVal    :: !CVal
              }
-        deriving (Eq, Ord)
+             deriving (Eq, Ord, G.Data)
 
 -- | A generalized CV allows for expressions involving infinite and epsilon values/intervals Used in optimization problems.
 data GeneralizedCV = ExtendedCV ExtCV
@@ -214,9 +217,8 @@ showExtCV = go False
                                        | True       = v ++ " :: "  ++ showBaseKind (kindOf extCV)
 
                  add :: String -> String -> String
-                 add n v
-                  | "-" `isPrefixOf` v = n ++ " - " ++ tail v
-                  | True               = n ++ " + " ++ v
+                 add n ('-':v) = n ++ " - " ++ v
+                 add n v       = n ++ " + " ++ v
 
                  mul :: String -> String -> String
                  mul n v = n ++ " * " ++ v
@@ -454,7 +456,7 @@ showCV shk w               = liftCV show show show show show show show show snd
               | True       = v
               where needsParen = case dropWhile isSpace v of
                                    []         -> False
-                                   rest@(x:_) -> any isSpace rest && x `notElem` "{[("
+                                   rest@(x:_) -> x == '-' || (any isSpace rest && x `notElem` "{[(")
 
 -- | Create a constant word from an integral.
 mkConstCV :: Integral a => Kind -> a -> CV
