@@ -16,7 +16,7 @@
 
 {-# OPTIONS_GHC -Wall -Werror #-}
 
-module Data.SBV.SMT.SMTLib2(cvt, cvtExp, cvtInc, declUserFuns) where
+module Data.SBV.SMT.SMTLib2(cvt, cvtExp, cvtCV, cvtInc, declUserFuns, constructTables) where
 
 import Data.Bits  (bit)
 import Data.List  (intercalate, partition, nub, elemIndex)
@@ -47,8 +47,8 @@ tbd :: String -> a
 tbd e = error $ "SBV.SMTLib2: Not-yet-supported: " ++ e
 
 -- | Translate a problem into an SMTLib2 script
-cvt :: SMTLibConverter [String]
-cvt ctx curProgInfo kindInfo isSat comments allInputs (allConsts, consts) tbls arrs uis defs (SBVPgm asgnsSeq) cstrs out cfg = pgm
+cvt :: SMTLibConverter ([String], [String])
+cvt ctx curProgInfo kindInfo isSat comments allInputs (allConsts, consts) tbls arrs uis defs (SBVPgm asgnsSeq) cstrs out cfg = (pgm, exportedDefs)
   where hasInteger     = KUnbounded `Set.member` kindInfo
         hasReal        = KReal      `Set.member` kindInfo
         hasFP          =  not (null [() | KFP{} <- Set.toList kindInfo])
@@ -66,6 +66,7 @@ cvt ctx curProgInfo kindInfo isSat comments allInputs (allConsts, consts) tbls a
         hasArrayInits  = (not . null) $  [() | (_, (_, _, ArrayFree (Left  (Just _)))) <- arrs]
                                       ++ [() | (_, (_, _, ArrayFree (Right _       ))) <- arrs]
         hasOverflows   = (not . null) [() | (_ :: OvOp) <- G.universeBi asgnsSeq]
+        hasQuantBools  = (not . null) [() | QuantifiedBool _ <- G.universeBi asgnsSeq]
         hasList        = any isList kindInfo
         hasSets        = any isSet kindInfo
         hasTuples      = not . null $ tupleArities
@@ -154,6 +155,7 @@ cvt ctx curProgInfo kindInfo isSat comments allInputs (allConsts, consts) tbls a
            | hasRegExp             = setAll "has regular expressions"
            | hasArrayInits         = setAll "has array initializers"
            | hasOverflows          = setAll "has overflow checks"
+           | hasQuantBools         = setAll "has quantified booleans"
 
            | hasFP || hasRounding
            = if needsQuantifiers
@@ -237,7 +239,7 @@ cvt ctx curProgInfo kindInfo isSat comments allInputs (allConsts, consts) tbls a
              ++ [ "; --- SBV Function definitions" | not (null funcMap) ]
              ++ concat [ declSBVFunc op nm | (op, nm) <- M.toAscList funcMap ]
              ++ [ "; --- user defined functions ---"]
-             ++ declUserFuns defs
+             ++ userDefs
              ++ [ "; --- assignments ---" ]
              ++ concatMap (declDef curProgInfo cfg tableMap funcMap) asgns
              ++ [ "; --- arrayDelayeds ---" ]
@@ -249,9 +251,15 @@ cvt ctx curProgInfo kindInfo isSat comments allInputs (allConsts, consts) tbls a
              ++ [ "; --- formula ---" ]
              ++ finalAssert
 
-        allTables      = [(t, genTableData rm (map fst consts) t) | t <- tbls]
-        constTables    = [(t, d) | (t, Left  d) <- allTables]
-        nonConstTables = [(t, d) | (t, Right d) <- allTables]
+        userDefs     = declUserFuns defs
+        exportedDefs
+          | null userDefs
+          = ["; No calls to 'smtFunction' found."]
+          | True
+          = "; Automatically generated by SBV. Do not modify!" : userDefs
+
+
+        (tableMap, constTables, nonConstTables) = constructTables rm consts tbls
 
         (arrayConstants, arrayDelayeds, arraySetups) = unzip3 $ map (declArray cfg allConsts) arrs
         delayedEqualities = concatMap snd nonConstTables
@@ -291,9 +299,6 @@ cvt ctx curProgInfo kindInfo isSat comments allInputs (allConsts, consts) tbls a
                   | s == falseSV = Just $ Left falseSV
                   | True         = Just $ Left s
 
-        tableMap  = IM.fromList $ map grab allTables
-          where grab (((t, _, _), _), _) = (t, "table" ++ show t)
-
         -- SBV only functions.
         funcMap = M.fromList reverses
           where reverses = zip (nub [op | op@(SeqOp SBVReverse{}) <- G.universeBi asgnsSeq])
@@ -501,11 +506,10 @@ cvtInc curProgInfo inps newKs (allConsts, consts) arrs tbls uis (SBVPgm asgnsSeq
 
         (arrayConstants, arrayDelayeds, arraySetups) = unzip3 $ map (declArray cfg allConsts) arrs
 
-        allTables = [(t, either id id (genTableData rm (map fst consts) t)) | t <- tbls]
-        (tableDecls, tableAssigns) = unzip $ map mkTable allTables
+        (tableMap, allTables) = (tm, ct ++ nct)
+            where (tm, ct, nct) = constructTables rm consts tbls
 
-        tableMap  = IM.fromList $ map mkIdx allTables
-          where mkIdx (((t, _, _), _), _) = (t, "table" ++ show t)
+        (tableDecls, tableAssigns) = unzip $ map mkTable allTables
 
         -- If we need flattening in models, do emit the required lines if preset
         settings
@@ -551,8 +555,8 @@ mkRelEq nm (fun, order) ak = res
          tk  = smtType ak
          res = "(forall ((x " ++ tk ++ ") (y " ++ tk ++ ")) (= " ++ lhs ++ " " ++ rhs ++ "))"
 
-declUI :: ProgInfo -> (String, (Maybe [String], SBVType)) -> [String]
-declUI ProgInfo{progTransClosures} (i, (_, t)) = declareName i t Nothing ++ declClosure
+declUI :: ProgInfo -> (String, (Bool, Maybe [String], SBVType)) -> [String]
+declUI ProgInfo{progTransClosures} (i, (_, _, t)) = declareName i t Nothing ++ declClosure
   where declClosure | Just external <- lookup i progTransClosures
                     =  declareName external t Nothing
                     ++ ["(assert " ++ mkRelEq external ("transitive-closure", i) (argKind t) ++ ")"]
@@ -564,28 +568,28 @@ declUI ProgInfo{progTransClosures} (i, (_, t)) = declareName i t Nothing ++ decl
 
 -- Note that even though we get all user defined-functions here (i.e., lambda and axiom), we can only have defined-functions
 -- and axioms. We spit axioms as is; and topologically sort the definitions.
-declUserFuns :: [SMTDef] -> [String]
+declUserFuns :: [(SMTDef, SBVType)] -> [String]
 declUserFuns ds
   | not (null lambdas)
   = error $ unlines $ "Data.SBV.declUserFuns: Unexpected anonymous functions in declDef:" : map show lambdas
   | True
   = declFuncs others
-  where (lambdas, others) = partition isLam ds
+  where (lambdas, others) = partition (isLam . fst) ds
         isLam SMTLam{} = True
         isLam SMTDef{} = False
 
 -- We need to topologically sort the user given definitions and axioms and put them in the proper order and construct.
 -- Note that there are no anonymous functions at this level.
-declFuncs :: [SMTDef] -> [String]
+declFuncs :: [(SMTDef, SBVType)] -> [String]
 declFuncs ds = map declGroup sorted
   where mkNode d = (d, getKey d, getDeps d)
 
-        getKey d = case d of
-                     SMTDef n _ _ _ _ -> n
-                     SMTLam{}         -> error $ "Data.SBV.declFuns: Unexpected definition kind: " ++ show d
+        getKey (d, _) = case d of
+                         SMTDef n _ _ _ _ -> n
+                         SMTLam{}         -> error $ "Data.SBV.declFuns: Unexpected definition kind: " ++ show d
 
-        getDeps (SMTDef _ _ d _ _) = d
-        getDeps l@SMTLam{}         = error $ "Data.SBV.declFuns: Unexpected definition kind: " ++ show l
+        getDeps (SMTDef _ _ d _ _, _) = d
+        getDeps (l@SMTLam{}, t)       = error $ "Data.SBV.declFuns: Unexpected definition: " ++ show (l, t)
 
         mkDecl Nothing  rt = "() "    ++ rt
         mkDecl (Just p) rt = p ++ " " ++ rt
@@ -598,8 +602,8 @@ declFuncs ds = map declGroup sorted
                                          [x] -> declUserDef True x
                                          xs  -> declUserDefMulti xs
 
-        declUserDef _ d@SMTLam{} = error $ "Data.SBV.declFuns: Unexpected anonymous lambda in user-defined functions: " ++ show d
-        declUserDef isRec (SMTDef nm fk deps param body) = ("; -- user given definition: " ++ nm ++ recursive ++ frees ++ "\n") ++ s
+        declUserDef _ d@(SMTLam{}, _) = error $ "Data.SBV.declFuns: Unexpected anonymous lambda in user-defined functions: " ++ show d
+        declUserDef isRec (SMTDef nm fk deps param body, ty) = ("; " ++ nm ++ " :: " ++ show ty ++ recursive ++ frees ++ "\n") ++ s
            where (recursive, definer) | isRec = (" [Recursive]", "define-fun-rec")
                                       | True  = ("",             "define-fun")
 
@@ -613,12 +617,12 @@ declFuncs ds = map declGroup sorted
 
         -- declare a bunch of mutually-recursive functions
         declUserDefMulti bs = render $ map collect bs
-          where collect d@SMTLam{} = error $ "Data.SBV.declFuns: Unexpected lambda in user-defined mutual-recursion group: " ++ show d
-                collect (SMTDef nm fk deps param body) = (deps, nm, '(' : nm ++ " " ++  decl ++ ")", body 3)
+          where collect d@(SMTLam{}, _) = error $ "Data.SBV.declFuns: Unexpected lambda in user-defined mutual-recursion group: " ++ show d
+                collect (SMTDef nm fk deps param body, ty) = (deps, nm, ty, '(' : nm ++ " " ++  decl ++ ")", body 3)
                   where decl = mkDecl param (smtType fk)
 
                 render defs = intercalate "\n" $
-                                  [ "; -- user given mutually-recursive definitions: " ++ intercalate ", " [n | (_, n, _, _) <- defs]
+                                  [ "; " ++ intercalate ", " [n ++ " :: " ++ show ty | (_, n, ty, _, _) <- defs]
                                   , "(define-funs-rec"
                                   ]
                                ++ [ open i ++ param d ++ close1 i | (i, d) <- zip [1..] defs]
@@ -626,10 +630,10 @@ declFuncs ds = map declGroup sorted
                      where open 1 = "  ("
                            open _ = "   "
 
-                           param (_, _, p, _) = p
+                           param (_deps, _nm, _ty, p, _body) = p
 
-                           dump (deps, nm, _, body) = "; Definition of: " ++ nm ++ ". [Refers to: " ++ intercalate ", " deps ++ "]"
-                                                    ++ "\n" ++ body
+                           dump (deps, nm, ty, _, body) = "; Definition of: " ++ nm ++ " :: " ++ show ty ++ ". [Refers to: " ++ intercalate ", " deps ++ "]"
+                                                        ++ "\n" ++ body
 
                            ld = length defs
 
@@ -663,6 +667,19 @@ nonConstTable (((i, ak, rk), _elts), _) = decl
   where t    = "table" ++ show i
         decl = "(declare-fun " ++ t ++ " (" ++ smtType ak ++ ") " ++ smtType rk ++ ")"
 
+constructTables :: RoundingMode -> [(SV, CV)] -> [((Int, Kind, Kind), [SV])]
+                -> ( IM.IntMap String                            -- table enumeration
+                   , [(((Int, Kind, Kind), [SV]), [String])]     -- constant tables
+                   , [(((Int, Kind, Kind), [SV]), [String])]     -- non-constant tables
+                   )
+constructTables rm consts tbls = (tableMap, constTables, nonConstTables)
+ where allTables      = [(t, genTableData rm (map fst consts) t) | t <- tbls]
+       constTables    = [(t, d) | (t, Left  d) <- allTables]
+       nonConstTables = [(t, d) | (t, Right d) <- allTables]
+       tableMap       = IM.fromList $ map grab allTables
+
+       grab (((t, _, _), _), _) = (t, "table" ++ show t)
+
 -- Left if all constants, Right if otherwise
 genTableData :: RoundingMode -> [SV] -> ((Int, Kind, Kind), [SV]) -> Either [String] [String]
 genTableData rm consts ((i, aknd, _), elts)
@@ -780,6 +797,9 @@ cvtExp curProgInfo caps rm tableMap functionMap expr@(SBVApp _ arguments) = sh e
 
         addRM s = s ++ " " ++ smtRoundingMode rm
 
+        hd _ (a:_) = a
+        hd w []    = error $ "Impossible: " ++ w ++ ": Received empty list of args!"
+
         -- lift a binary op
         lift2  o _ [x, y] = "(" ++ o ++ " " ++ x ++ " " ++ y ++ ")"
         lift2  o _ sbvs   = error $ "SBV.SMTLib2.sh.lift2: Unexpected arguments: "   ++ show (o, sbvs)
@@ -796,13 +816,14 @@ cvtExp curProgInfo caps rm tableMap functionMap expr@(SBVApp _ arguments) = sh e
 
         liftAbs sgned args | fpOp        = lift1 "fp.abs" sgned args
                            | intOp       = lift1 "abs"    sgned args
-                           | bvOp, sgned = mkAbs (head args) "bvslt" "bvneg"
-                           | bvOp        = head args
-                           | True        = mkAbs (head args) "<"     "-"
-          where mkAbs x cmp neg = "(ite " ++ ltz ++ " " ++ nx ++ " " ++ x ++ ")"
+                           | bvOp, sgned = mkAbs fArg "bvslt" "bvneg"
+                           | bvOp        = fArg
+                           | True        = mkAbs fArg "<"     "-"
+          where fArg = hd "liftAbs" args
+                mkAbs x cmp neg = "(ite " ++ ltz ++ " " ++ nx ++ " " ++ x ++ ")"
                   where ltz = "(" ++ cmp ++ " " ++ x ++ " " ++ z ++ ")"
                         nx  = "(" ++ neg ++ " " ++ x ++ ")"
-                        z   = cvtCV rm (mkConstCV (kindOf (head arguments)) (0::Integer))
+                        z   = cvtCV rm (mkConstCV (kindOf (hd "liftAbs.arguments" arguments)) (0::Integer))
 
         lift2B bOp vOp
           | boolOp = lift2 bOp
@@ -837,7 +858,7 @@ cvtExp curProgInfo caps rm tableMap functionMap expr@(SBVApp _ arguments) = sh e
                       | True = lift2 o
 
         unintComp o [a, b]
-          | KUserSort s (Just _) <- kindOf (head arguments)
+          | KUserSort s (Just _) <- kindOf (hd "unintComp" arguments)
           = let idx v = "(" ++ s ++ "_constrIndex " ++ v ++ ")" in "(" ++ o ++ " " ++ idx a ++ " " ++ idx b ++ ")"
         unintComp o sbvs = error $ "SBV.SMT.SMTLib2.sh.unintComp: Unexpected arguments: "   ++ show (o, sbvs, map kindOf arguments)
 
@@ -845,7 +866,7 @@ cvtExp curProgInfo caps rm tableMap functionMap expr@(SBVApp _ arguments) = sh e
         stringOrChar KChar   = True
         stringOrChar _       = False
         stringCmp swap o [a, b]
-          | stringOrChar (kindOf (head arguments))
+          | stringOrChar (kindOf (hd "stringCmp" arguments))
           = let (a1, a2) | swap = (b, a)
                          | True = (a, b)
             in "(" ++ o ++ " " ++ a1 ++ " " ++ a2 ++ ")"
@@ -853,7 +874,7 @@ cvtExp curProgInfo caps rm tableMap functionMap expr@(SBVApp _ arguments) = sh e
 
         -- NB. Likewise for sequences
         seqCmp swap o [a, b]
-          | KList{} <- kindOf (head arguments)
+          | KList{} <- kindOf (hd "seqCmp" arguments)
           = let (a1, a2) | swap = (b, a)
                          | True = (a, b)
             in "(" ++ o ++ " " ++ a1 ++ " " ++ a2 ++ ")"
@@ -1010,8 +1031,7 @@ cvtExp curProgInfo caps rm tableMap functionMap expr@(SBVApp _ arguments) = sh e
           | True  = reducePB pb args'
           where args' = map cvtSV args
 
-        -- NB: Z3 semantics have the predicates reversed: i.e., it returns true if overflow isn't possible. Hence the not.
-        sh (SBVApp (OverflowOp op) args) = "(not (" ++ show op ++ " " ++ unwords (map cvtSV args) ++ "))"
+        sh (SBVApp (OverflowOp op) args) = "(" ++ show op ++ " " ++ unwords (map cvtSV args) ++ ")"
 
         -- Note the unfortunate reversal in StrInRe..
         sh (SBVApp (StrOp (StrInRe r)) args) = "(str.in.re " ++ unwords (map cvtSV args) ++ " " ++ regExpToSMTString r ++ ")"
@@ -1039,7 +1059,6 @@ cvtExp curProgInfo caps rm tableMap functionMap expr@(SBVApp _ arguments) = sh e
         sh (SBVApp (SetOp SetSubset)     args)   = "(subset "       ++ unwords (map cvtSV args) ++ ")"
         sh (SBVApp (SetOp SetDifference) args)   = "(setminus "     ++ unwords (map cvtSV args) ++ ")"
         sh (SBVApp (SetOp SetComplement) args)   = "(complement "   ++ unwords (map cvtSV args) ++ ")"
-        sh (SBVApp (SetOp SetHasSize)    args)   = "(set-has-size " ++ unwords (map cvtSV args) ++ ")"
 
         sh (SBVApp (TupleConstructor 0)   [])    = "mkSBVTuple0"
         sh (SBVApp (TupleConstructor n)   args)  = "((as mkSBVTuple" ++ show n ++ " " ++ smtType (KTuple (map kindOf args)) ++ ") " ++ unwords (map cvtSV args) ++ ")"
@@ -1082,7 +1101,7 @@ cvtExp curProgInfo caps rm tableMap functionMap expr@(SBVApp _ arguments) = sh e
           | Just f <- lookup op uninterpretedTable
           = f (map cvtSV args)
           | True
-          = if not (null args) && isUserSort (head args)
+          = if not (null args) && isUserSort (hd "isUserSort" args)
             then error $ unlines [ ""
                                  , "*** Cannot translate operator        : " ++ show op
                                  , "*** When applied to arguments of kind: " ++ intercalate ", " (nub (map (show . kindOf) args))
