@@ -82,9 +82,11 @@ haskell vname vs = intercalate "\n" $ [ "-- Automatically generated by SBV. Do n
                                    ++ [ n ++ " :: " ++ getType vs
                                       , n ++ " = [ " ++ intercalate ("\n" ++ pad ++  ", ") (map mkLine vs), pad ++ "]"
                                       ]
-  where n | null vname                 = "testVectors"
-          | not (isAlpha (head vname)) = "tv" ++ vname
-          | True                       = vname
+  where n = case vname of
+              ""                    -> "testVectors"
+              f:_ | not (isAlpha f) -> "tv" ++ vname
+                  | True            -> vname
+
         imports
           | null vs               = []
           | needsInt && needsWord = ["import Data.Int", "import Data.Word", ""]
@@ -195,11 +197,15 @@ c n vs = intercalate "\n" $
               , "typedef struct {"
               , "  struct {"
               ]
-           ++ (if null vs then [] else zipWith (mkField "i") (fst (head vs)) [(0::Int)..])
+           ++ (case vs of
+                 []       -> []
+                 (i, _):_ -> zipWith (mkField "i") i [(0::Int)..])
            ++ [ "  } input;"
               , "  struct {"
               ]
-           ++ (if null vs then [] else zipWith (mkField "o") (snd (head vs)) [(0::Int)..])
+           ++ (case vs of
+                 []       -> []
+                 (_, o):_ -> zipWith (mkField "o") o [(0::Int)..])
            ++ [ "  } output;"
               , "} " ++ n ++ "TestVector;"
               , ""
@@ -275,7 +281,10 @@ c n vs = intercalate "\n" $
           | True    = "printf(\"%*d. " ++ fmtString ++ "\\n\", " ++ show (length (show (length vs - 1))) ++ ", i"
                     ++ concatMap ("\n           , " ++ ) (zipWith inp is [(0::Int)..] ++ zipWith out os [(0::Int)..])
                     ++ ");"
-          where (is, os) = head vs
+          where (is, os) = case vs of
+                             h:_ -> h
+                             _   -> error "outLine: Impossible hapepned, empty vs!"
+
                 inp cv i = mkBool cv (n ++ "[i].input.i"  ++ show i)
                 out cv i = mkBool cv (n ++ "[i].output.o" ++ show i)
                 mkBool cv s = case kindOf cv of
@@ -309,14 +318,18 @@ forte vname bigEndian ss vs = intercalate "\n" $ [ "// Automatically generated b
                                           ++ [ "   in [ " ++ intercalate "\n      , " (map mkLine vs)
                                              , "      ];"
                                              ]
-  where n | null vname                 = "testVectors"
-          | not (isAlpha (head vname)) = "tv" ++ vname
-          | True                       = vname
+  where n = case vname of
+              ""                    -> "testVectors"
+              f:_ | not (isAlpha f) -> "tv" ++ vname
+                  | True            -> vname
+
         blaster
          | bigEndian = "map (\\s. s == \"1\") (explode (string_tl r))"
          | True      = "rev (map (\\s. s == \"1\") (explode (string_tl r)))"
+
         toF True  = '1'
         toF False = '0'
+
         blast cv = let noForte w = error "SBV.renderTest: " ++ w ++ " values are not supported when generating Forte test-cases."
                    in case kindOf cv of
                         KBool             -> [toF (cvToBool cv)]
